{
  "type": "record",
  "name": "TelemetryKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for the telemetry client ID, only one should ever exist",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "The telemetry entity name, which serves as a unique id"
    }
  ],
  "Aspect": {
    "name": "telemetryKey",
    "keyForEntity": "telemetry",
    "entityCategory": "internal",
    "entityAspects": [
      "telemetryClientId"
    ]
  }
}